"""
Amatino API Python Bindings
Amatino Error Module
Author: hugh@amatino.io
"""


class AmatinoError(Exception):
    """
    Base exception class for all errors thrown by the Amatino library
    """
    pass
